
#IFDEF NOT __GCI_INC__ THEN
#DEFINE __GCI_INC__

#DEFINE _GCI_LINSEC_OFFSET 315532800 

#DEFINE GCI_MESSAGE_TYPE_UNKNOWN   -1
#DEFINE GCI_MESSAGE_TYPE_METHOD     2
#DEFINE GCI_MESSAGE_TYPE_CONFIG     6

VAR
   _gci_private_keepalive : INT := 120;
   
   _gci_private_deviceId:STRING;
   _gci_private_project_id  : STRING;
   _gci_private_cert        : STRING;
   _gci_private_cert_pass   : STRING;
   
   _gci_private_config_topic : STRING;

   _gci_private_dm_topic: STRING;
   _gci_private_dm_method:STRING;
   _gci_private_dm_active:BOOL := FALSE;
   
   _gci_private_mqtt_rxd  : mqttReceive;
   _gci_private_mqtt_handle: INT := -1;
   
   _gci_private_tz_offset : DINT;
   _gci_private_lifetime: DINT;
   
   _gci_private_token_refresh_time : DINT;
   _gci_private_connection_lost : BOOL := TRUE;
   
END_VAR;



FUNCTION _gciParseMethod:INT;
VAR_INPUT
   str:STRING;
END_VAR;
VAR
   i:INT;
END_VAR;
   
   _gci_private_dm_method := strToken(str:=str, delimiter:="/", index:=2);
   //DebugFmt(message:="parse method: "+str+" => "+_gci_private_dm_method);
  
   _gci_private_dm_active := TRUE;
END_FUNCTION;



FUNCTION _gciSetupVars: INT;
VAR_INPUT
   project_id : STRING;
   region_id : STRING;
   registry_id: STRING;
   device_id: STRING;
   o_clientid: ACCESS STRING;
END_VAR;
   _gci_private_deviceId := device_id;
   
   o_clientid := "projects/" + project_id + "/locations/" + region_id + "/registries/" + registry_id + "/devices/"+device_id;
    
   _gci_private_dm_topic := "/devices/"+device_id+"/commands";
   
   _gci_private_config_topic := "/devices/"+device_id+"/config";
END_FUNCTION;

FUNCTION _clock_now:DINT;
   _clock_now := clockNow()+_GCI_LINSEC_OFFSET - _gci_private_tz_offset;// Subtract TZ
END_FUNCTION;


FUNCTION _cloudGenJWT: STRING;
VAR
   jwt: SYSHANDLE;
   str: STRING := "";
   rc : INT;
   now : DINT;
END_VAR;

   // ********************************************
   // *                                          *
   // *     JWT requires Large String Support    *
   // *                                          *
   // ********************************************
   rc := jwtCreate(jwt:=jwt);

   //DebugFmt(message:="jwtCreate: \1", v1:=rc);

   IF rc = 1 THEN
      now := _clock_now();
      
      rc := jwtClaimAddDint(jwt:=jwt, name:="iat", value:= now );
      //DebugFmt(message:="jwtClaimAddInt: \1", v1:=rc);
      
      rc := jwtClaimAddDint(jwt:=jwt, name:="exp", value:=now + _gci_private_lifetime);
      //DebugFmt(message:="jwtClaimAddInt: \1", v1:=rc);
      
      
      rc := jwtClaimAdd(jwt:=jwt, name:="aud", value:=_gci_private_project_id);
      //DebugFmt(message:="jwtClaimAdd: \1", v1:=rc);
         
      // Use RS256
      rc := jwtAlgSetAsym(jwt:=jwt, type := 4, cert := _gci_private_cert, pass := _gci_private_cert_pass);
      //DebugFmt(message:="jwtSetAlgSym: \1", v1:=rc);

      rc := jwtAlgGet(jwt:=jwt);
      //DebugFmt(message:="jwtGetAlg: \1", v1:=rc);

      rc := jwtEncode(jwt:=jwt, token:=str);
      //DebugFmt(message:="jwtEncode: \1, "+str, v1:=rc);
      
      jwtFree(jwt:=jwt);
      
      // Determine when to refresh the credentials. 
      // The actual expiration time is 10 minutes later that specified to account for clock skew,
      // so we refresh it when it expires.
      _gci_private_token_refresh_time := now + _gci_private_lifetime;
      
      //DebugFmt(message:="new refresh time: \4", v4:=_gci_private_token_refresh_time );
      
      _cloudGenJWT := str;
   ELSE
      _cloudGenJWT := "";
   END_IF;

END_FUNCTION;



///////////////////////////////////////////////////////////////////////////////////////////////////////
// _gciSubscribe()
//
// Description:
//    Subscribe to the events on the connected IoT core.
//
// Input:
//    None.
//
// Returns:
//    0         - Success.
//    1         - Invalid connection.
//    2         - Not connected to server.
//    3         - Invalid parameter
//
FUNCTION ALIGN _gciSubscribe : INT;
   IF(_gci_private_mqtt_handle < 0) THEN
      _gciSubscribe := 1;
      RETURN;
   END_IF;
   _gciSubscribe := mqttSubscribe(handle := _gci_private_mqtt_handle, qos := 1, topic := _gci_private_config_topic);
   IF(_gciSubscribe <> 0) THEN 
      RETURN;
   END_IF;

   _gciSubscribe := mqttSubscribe(handle := _gci_private_mqtt_handle, qos := 1, topic := _gci_private_dm_topic+"/#");
   IF(_gciSubscribe <> 0) THEN 
      RETURN;
   END_IF;
   
END_FUNCTION;



///////////////////////////////////////////////////////////////////////////////////////////////////////
// gciOpenJWT()
//
// Description:
//    Open the connection to the IoT core using a JSON Web Token for authentication.
//
// Input:
//    project_id  - The project identifier.
//    region_id   - The region identifier. This is the region where the server is located.
//    registry_id - The id of the registry.
//    device_id   - The device identifier.
//    cert        - The name of the certificate containing the private key.
//    cert_pass   - The password for the private key, if needed
//    server      - The hostname of the IoT core.
//    iface       - The network interface to use.
//
// Returns:
//    0          - Success.
//    -1         - No available connections.
//    -2         - Invalid parameter
//    -3         - Connection is already in use.
//    -4         - Failed to create token.
//
FUNCTION ALIGN gciOpenJWT : INT;
VAR_INPUT
   project_id  : MANDATORY STRING;
   region_id   : MANDATORY STRING;
   registry_id : MANDATORY STRING;
   device_id   : MANDATORY STRING;
   
   server      : STRING := "mqtt.googleapis.com";
   port        : DINT := 8883;
   iface       : SINT;
   
   cert        : MANDATORY STRING;
   cert_pass   : STRING;
   
   lifetime    : DINT := 86400;// 24 hours
   tz_offset   : DINT := 0;
END_VAR;
VAR
   token       : STRING;
   clientid    : STRING;
END_VAR;

   
   IF _gci_private_mqtt_handle >= 0 THEN
      gciOpenJWT := -3;
      RETURN;
   END_IF;
   
   IF lifetime > 24*60*60 OR lifetime < 0 THEN
      gciOpenJWT := -2;
      RETURN;
   END_IF;

   _gci_private_tz_offset := tz_offset;
   _gci_private_lifetime := lifetime;
   _gci_private_project_id := project_id;
   _gci_private_cert:=cert;
   _gci_private_cert_pass:=cert_pass;
   
   _gciSetupVars(project_id:=project_id, region_id:=region_id, registry_id:=registry_id, device_id:=device_id, o_clientid:=clientid);

   token := _cloudGenJWT();

   IF token = "" THEN
      gciOpenJWT := -4;
      RETURN;
   END_IF;

   _gci_private_mqtt_handle  := mqttOpen(ip := server, port := port, clientid := clientid, 
                                iface:=iface, useTLS:=TRUE, username := "unused", 
                                password := token, keepalive := _gci_private_keepalive, clean := FALSE);
   
   //DebugFmt(message:="Open JWT: \1, s: "+server+":\4, cl: "+clientid, v1:=_gci_private_mqtt_handle, v4:=port);
   IF(_gci_private_mqtt_handle < 0) THEN
      gciOpenJWT:=_gci_private_mqtt_handle;
   ELSE
      gciOpenJWT:=0;
   END_IF;
   
END_FUNCTION;




///////////////////////////////////////////////////////////////////////////////////////////////////////
// gciClose()
//
// Description:
//    Close the connection to the IoT core
//
// Input:
//    None.
//
// Returns:
//    None.
//
FUNCTION ALIGN gciClose;
   IF(_mqtt_handle >= 0) THEN
      mqttClose(handle:=_mqtt_handle);
   END_IF;
   _mqtt_handle := -1;
END_FUNCTION;



///////////////////////////////////////////////////////////////////////////////////////////////////////
// gciConnected()
//
// Description:
//    Tells if the connection to the IoT core is established.
//
// Input:
//    None.
//
// Returns:
//    TRUE      - Device is connected to IoT core.
//    FALSE     - Device is not connected to IoT core.
//
FUNCTION ALIGN gciConnected : BOOL;
   gciConnected := mqttConnected(handle := _gci_private_mqtt_handle);
END_FUNCTION;


///////////////////////////////////////////////////////////////////////////////////////////////////////
// gciStatus()
//
// Description:
//    Returns the status of the connection.
//
// Input:
//    None.
//
// Returns:
//    0         - Connected to the IoT core.
//    1         - Invalid connection.
//    2         - Not connected to IoT core.
//    3         - IoT core server not found.
//    4         - No reply from IoT core
//    5         - Connection rejected, unacceptable protocol version.
//    6         - Connection rejected, client ID rejected.
//    7         - Connection rejected, server unavailable.
//    8         - Connection rejected, bad user-name or password.
//    9         - Connection rejected, not authorized.
//    20        - Secure connection failed.
//    21        - Secure connection require client certificate.
//    22        - Certificate verifications failed.
//
FUNCTION ALIGN gciStatus : SINT;
   gciStatus := mqttStatus(handle := _gci_private_mqtt_handle);
END_FUNCTION;




///////////////////////////////////////////////////////////////////////////////////////////////////////
// gciWaitEvent()
//
// Description:
//    Waits for a new message, blocking the thread.
//
// Input:
//    timeout   - The number of seconds to wait before timing out. Use 0 to return immediately, use -1 to wait forever.
//
// Returns:
//    1         - A messages is received. Use gciReceiveMessage to read it.
//    0         - Timeout.
//   -1         - No connection.
//
FUNCTION ALIGN gciWaitEvent : INT;
VAR_INPUT
   timeout   : INT;
END_VAR;
VAR
   token : STRING;
   rc : INT;
END_VAR;
   IF (_gci_private_token_refresh_time - _clock_now()) <= 0 THEN
      // token has expired, generate a new one
      token := _cloudGenJWT();
      IF token <> "" THEN
         rc := mqttCredentialsSet(handle:=_gci_private_mqtt_handle, password := token, reconnect :=TRUE);
         //DebugFmt(message:="Set credentials: \1 => "+ token, v1:=rc);
      END_IF;
   END_IF;
   
   // Detect connection loss
  IF mqttConnectionLoss(handle := _gci_private_mqtt_handle) THEN
    //DebugMsg(message := "Connection to MQTT server has been lost");
    _gci_private_connection_lost := TRUE;
  END_IF;
  
  
  IF _gci_private_connection_lost AND mqttConnected(handle := _gci_private_mqtt_handle) THEN
     // reconnected - subscribe in case the subscriptions have been lost.
     _gciSubscribe();
     _gci_private_connection_lost := FALSE;
  END_IF;
  
   gciWaitEvent := mqttWaitEvent(timeout := timeout);
END_FUNCTION;

///////////////////////////////////////////////////////////////////////////////////////////////////////
// gciReceiveMessage()
//
// Description:
//    Reads a new message.
//
// Input:
//    data      - Address of a buffer to store the data of the message in.
//    maxsize   - The size of the buffer. Any message data after this limit is lost.
//
// Output:
//    ready     - True if a message has been received.
//    size      - The size of the read data.
//    type      - The type of the message:
//               -1  - Unknown message.
//                2  - Command message. Use gciGetMethodRequest to read the name of the requested command. The data contains the payload for the command.
//                6  - Config message. The data contains the received configuration.
//
//
FUNCTION_BLOCK ALIGN gciReceiveMessage;
VAR_INPUT
   data      : PTR;
   maxsize   : INT;
END_VAR;
VAR_OUTPUT
   size      : INT;
   ready     : BOOL;
   type      : INT;
END_VAR;
VAR
   rc:INT;
   rd : BOOL;
   typ : INT;
END_VAR;
   ready := FALSE;
   rc := mqttWaitEvent(timeout := 0);
   IF rc > 0 THEN
      //DebugFmt(message:="event: \1",v1:=rc);
      _gci_private_mqtt_rxd(data := (data), maxsize := maxsize);
      IF(_gci_private_mqtt_rxd.ready) THEN
         IF(_gci_private_mqtt_rxd.handle = _gci_private_mqtt_handle) THEN
            //DebugFmt(message:="Received "+_gci_private_mqtt_rxd.topic+", \1, test: "+_gci_private_config_topic+"/"+_gci_private_dm_topic, v1:=size);
            IF(strMid(str:=_gci_private_mqtt_rxd.topic, start:=1, length:=strLen(str:=_gci_private_config_topic)) = _gci_private_config_topic) THEN
               // Config

               type := GCI_MESSAGE_TYPE_CONFIG; 
            ELSIF (strMid(str:=_gci_private_mqtt_rxd.topic, start:=1, length:=strLen(str:=_gci_private_dm_topic)) = _gci_private_dm_topic) THEN
               // Direct method
               _gciParseMethod(str:=strMid(str:=_gci_private_mqtt_rxd.topic, start:=strLen(str:=_gci_private_dm_topic)+1));
               
               type := GCI_MESSAGE_TYPE_METHOD;
            ELSE
               // Unknown topic
               type := GCI_MESSAGE_TYPE_UNKNOWN;
            END_IF;
            size := _gci_private_mqtt_rxd.size;
            
            ready := TRUE;
            RETURN;
         END_IF;         
      END_IF;
   END_IF;
END_FUNCTION_BLOCK;






///////////////////////////////////////////////////////////////////////////////////////////////////////
// gciPublish()
//
// Description:
//    Sends a message to the connected IoT core.
//
// Input:
//    subfolder   - The subfolder to use as event catagory or classification.
//    data        - Address of a buffer with the data to send.
//    size        - The size of the data to send
//
// Returns:
//    0         - Success.
//    1         - Invalid connection.
//    2         - Not connected to server.
//    3         - Invalid parameter
//    4         - A publish is already in progress
//    5         - Message transmission interrupted due to connection loss. The message will be transmitted again when connection is reestablished.
//
FUNCTION ALIGN gciPublish : INT;
VAR_INPUT
   subfolder     : STRING;
   data      : PTR;
   size      : INT;
END_VAR;
VAR
   topic     :STRING;
END_VAR;
   topic  := "/devices/"+_gci_private_deviceId+"/events";
   IF subfolder <> "" THEN
      topic  := topic+"/"+subfolder;
   END_IF;
   gciPublish := mqttPublish(
                     handle   := _gci_private_mqtt_handle,
                     topic    := topic,
                     qos      := 1,
                     retained := OFF,
                     data     := data,
                     size     := size
                    );
     //DebugFmt(message:="publish: "+topic);
                       
END_FUNCTION;


///////////////////////////////////////////////////////////////////////////////////////////////////////
// gciGetMethodRequest()
//
// Description:
//    Retrieves the name of the requested method.
//
// Output:
//    method    - The name of the called method.
//    
// Returns:
//    0         - Success.
//   -1         - No unread method request available.
//
FUNCTION ALIGN gciGetMethodRequest:INT;
VAR_INPUT
   method:ACCESS STRING;
END_VAR;

   IF(_gci_private_dm_active) THEN
      method := _gci_private_dm_method;
      gciGetMethodRequest := 0;
      _gci_private_dm_active := FALSE;
   ELSE
      gciGetMethodRequest := -1;
   END_IF;

END_FUNCTION;



///////////////////////////////////////////////////////////////////////////////////////////////////////
// gciSetState()
//
// Description:
//    Sends the device state to the connected IoT core.
//
// Input:
//    data        - Address of a buffer with the state to send.
//    size        - The size of the state to send
//
// Returns:
//    0         - Success.
//    1         - Invalid connection.
//    2         - Not connected to server.
//    3         - Invalid parameter
//    4         - A publish is already in progress
//    5         - Message transmission interrupted due to connection loss. The message will be transmitted again when connection is reestablished.
//
FUNCTION ALIGN gciSetState: INT;
VAR_INPUT
   data      : PTR;
   size      : INT;
END_VAR;
VAR
   topic     :STRING;
END_VAR;
   topic  := "/devices/"+_gci_private_deviceId+"/state";
   
   gciSetState := mqttPublish(
                     handle   := _gci_private_mqtt_handle,
                     topic    := topic,
                     qos      := 1,
                     retained := OFF,
                     data     := data,
                     size     := size
                    );
     //DebugFmt(message:="publish: "+topic);
                       
END_FUNCTION;


#END_IF; /* __GCI_INC__ */ 